from airontools.devices import get_available_gpus


def test_get_available_devices():
    available_gpus = get_available_gpus()
    assert isinstance(available_gpus, list)
    assert all(isinstance(element, str) for element in available_gpus)



import warnings
from typing import Optional, Tuple, Union

import keras
import numpy as np
import tensorflow as tf

from airontools.constructors.utils import get_regularizer


def layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    units: int,
    name: str = "layer",
    name_ext: Optional[str] = None,
    num_heads: int = 0,
    key_dim: int = 0,
    value_dim: int = 0,
    multi_head_attention_dropout_rate: float = 0.0,
    return_attention_scores: bool = False,
    use_causal_mask: bool = False,
    activation: Union[str, keras.layers.Activation] = "linear",
    sequential_activation: Union[str] = "tanh",
    use_bias: bool = True,
    sequential: bool = False,
    bidirectional: bool = False,
    return_sequences: bool = False,
    filters: int = 0,
    kernel_size: Union[int, Tuple[int]] = 0,
    padding: str = "valid",
    pooling: Optional[Union[str, keras.layers.Layer]] = None,
    pool_size: Union[int, Tuple[int]] = 1,
    conv_transpose: bool = False,
    strides: Union[int, Tuple[int]] = 1,
    sequential_axis: int = 1,
    kernel_regularizer_l1: float = 0.001,
    kernel_regularizer_l2: float = 0.001,
    bias_regularizer_l1: float = 0.001,
    bias_regularizer_l2: float = 0.001,
    dropout_rate: float = 0.0,
    normalization_type: Optional[str] = None,
) -> Union[keras.layers.Layer, Tuple[keras.layers.Layer, keras.layers.Layer]]:
    """It builds a custom layer. For now only 2D convolutions
    are supported for input of rank 4.

    Parameters:
        x (tf.Tensor, keras.layers.Layer): Input layer or tensor.
        units (int): Number of units for the dense layer. If a value is given, a dense layer will be added
        automatically if not sequential, else a sequential model. Useful to force an output dimensionality of the
        custom layer when using convolutional layers.
        name (str): Name of the custom layer.
        name_ext (str): Extension name for the custom layer that will be at the end of it.
        num_heads (int): Number of heads for the multi-head attention layer.
        key_dim (int): Key dimensionality for the multi-head attention layer, if None then the number of units is
        used instead.
        value_dim (int): Value dimensionality for the multi-head attention layer, if None then key_dim is used
        instead.
        multi_head_attention_dropout_rate (float): Multi-head attention dropout rate.
        return_attention_scores (bool): Whether to return attention scores or not.
        use_causal_mask: Whether to use casual mask in the multi-head attention.
        activation (str, keras.layers.Activation): The activation function of the output of the last hidden layer.
        sequential_activation (str): The activation function of the output of the sequential hidden layer.
        use_bias (bool): Whether to sue bias or not.
        sequential (bool): Whether to consider a sequential custom layer or not. Sequential and self-attention
        (num_heads > 0) are not compatible.
        bidirectional (bool): Whether to consider bidirectional case or not (only active if sequential).
        names are not repeated.
        return_sequences (bool): Whether to return sequences or not (only active if sequential).
        filters (int): Number of filters to be used. If a value is given, a convolutional layer will be
        automatically added.
        kernel_size (int, Tuple[int]): Kernel size for the convolutional layer.
        conv_transpose (bool): Whether to use a transpose conv layer or not (only active if filters and
        kernel_size are set).
        strides (int, Tuple[int]): Strides for the conv layer (only active if filters and
        kernel_size are set).
        padding (str): Padding to be applied (only active if filters and
        kernel_size are set).
        pooling (str, keras.layers.Layer): Pooling type.
        pool_size (int, Tuple[int]): Pooling size.
        sequential_axis (int): The axis that defines the sequence. For sequential models is normally 1. For
        self-attention (num_heads > 0) and image-like inputs, the sequential axis is the channel axis (3 for 2D
        images and 4 for 3D images).
        kernel_regularizer_l1 (float): Kernel regularization using l1 penalization (Lasso).
        kernel_regularizer_l2 (float): Kernel regularization using l2 penalization (Ridge).
        bias_regularizer_l1 (float): Bias regularization using l1 penalization (Lasso).
        bias_regularizer_l2 (float): Bias regularization using l2 penalization (Ridge).
        dropout_rate (float): Dropout rate.
        normalization_type (str): If set, a normalization layer (BN or LN) will be added right before the output activation function.

    Returns:
        x (keras.layers.Layer | tuple(keras.layers.Layer, keras.layers.Layer)): A keras layer.
    """

    if num_heads > 0 and units is None and key_dim == 0:
        warnings.warn(
            "in order to use a multi-head attention layer either units or key_dim needs to be set",
        )

    # Convolution
    conv_condition = all(
        [conv_param != 0 for conv_param in [filters, kernel_size]],
    )
    if conv_condition:
        pooling_dim = _get_pooling_dim(x)
        if isinstance(strides, int) and pooling_dim > 0:
            _strides = tuple([strides] * pooling_dim)
        else:
            _strides = strides
        conv_kwargs = dict(
            use_bias=use_bias,
            filters=filters,
            kernel_size=kernel_size,
            strides=_strides,
            padding=padding,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
        )
        convolutional_layer_name = "".join([name, "convolution"])
        x = convolutional_layer_constructor(
            x,
            name=convolutional_layer_name,
            name_ext=name_ext,
            conv_transpose=conv_transpose,
            **conv_kwargs,
        )

    # Pooling
    if pooling is not None:
        pooling_dim = _get_pooling_dim(x)
        if isinstance(pool_size, int) and pooling_dim > 0:
            _pool_size = tuple([pool_size] * pooling_dim)
        else:
            _pool_size = pool_size
        pooling_kwargs = dict(
            pool_size=_pool_size,
            strides=strides,
            padding=padding,
        )
        pooling_layer_name = "".join([name, "pooling"])
        x = pooling_layer_constructor(
            x,
            name=pooling_layer_name,
            name_ext=name_ext,
            pooling=pooling,
            **pooling_kwargs,
        )

    # Multi-Head Attention
    attention_scores = None
    if num_heads > 0:
        _key_dim = key_dim if key_dim > 0 else units
        _value_dim = value_dim if value_dim > 0 else _key_dim
        multi_head_attention_kwargs = dict(
            num_heads=num_heads,
            key_dim=_key_dim,
            value_dim=_value_dim,
            use_bias=use_bias,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
            dropout=multi_head_attention_dropout_rate,
        )
        self_attention_layer_name = "".join([name, "self_attention"])
        x = self_attention_layer_constructor(
            x,
            name=self_attention_layer_name,
            name_ext=name_ext,
            sequential_axis=sequential_axis,
            return_attention_scores=return_attention_scores,
            use_causal_mask=use_causal_mask,
            **multi_head_attention_kwargs,
        )
        if return_attention_scores:
            x, attention_scores = x

    # Sequential
    elif sequential:
        seq_kwargs = dict(
            units=units,
            use_bias=use_bias,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
            return_sequences=return_sequences,
            activation=sequential_activation,
        )
        sequential_layer_name = "".join([name, "sequential"])
        x = sequential_layer_constructor(
            x,
            name=sequential_layer_name,
            name_ext=name_ext,
            bidirectional=bidirectional,
            sequential_axis=sequential_axis,
            **seq_kwargs,
        )

    # Dense
    if units > 0:
        dense_kwargs = dict(
            units=units,
            use_bias=use_bias,
            kernel_regularizer=get_regularizer(
                kernel_regularizer_l1,
                kernel_regularizer_l2,
            ),
            bias_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
        )
        dense_layer_name = "".join([name, "dense"])
        x = dense_layer_constructor(
            x,
            name=dense_layer_name,
            name_ext=name_ext,
            return_sequences=return_sequences,
            **dense_kwargs,
        )

    # Normalization
    if normalization_type is None:
        pass
    elif normalization_type == "bn":
        bn_layer_name = "".join([name, "bn"])
        x = bn_layer_constructor(x, name=bn_layer_name, name_ext=name_ext)
    elif normalization_type == "ln":
        ln_layer_name = "".join([name, "ln"])
        x = ln_layer_constructor(x, name=ln_layer_name, name_ext=name_ext)
    else:
        raise ValueError(
            f"Unknown normalization type {normalization_type}. Only 'bn' and 'ln' are supported."
        )

    # Dropout
    if dropout_rate != 0:
        dropout_layer_name = "".join([name, "dropout"])
        x = dropout_layer_constructor(
            x,
            name=dropout_layer_name,
            name_ext=name_ext,
            dropout_rate=dropout_rate,
        )

    # Activation
    activation_kwargs = dict(
        alpha_regularizer=get_regularizer(bias_regularizer_l1, bias_regularizer_l2),
    )
    activation_layer_name = "".join([name, "activation"])
    x = activation_layer_constructor(
        x,
        name=activation_layer_name,
        name_ext=name_ext,
        activation=activation,
        **activation_kwargs,
    )

    if return_attention_scores:
        return x, attention_scores
    else:
        return x


def dropout_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    dropout_rate: float,
    name: str = "dropout",
    name_ext: Optional[str] = None,
) -> keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2 and all([shape is not None for shape in input_shape[1:]]):
        output_reshape = input_shape
        flatten_layer_name = "_".join([name, "pre", "dropout", "flatten"])
        if name_ext is not None:
            flatten_layer_name = "_".join([flatten_layer_name, name_ext])
        x = keras.layers.Flatten(name=flatten_layer_name)(x)
    dropout_layer_name = "_".join([name, "dropout"])
    if name_ext is not None:
        dropout_layer_name = "_".join([dropout_layer_name, name_ext])
    x = keras.layers.Dropout(name=dropout_layer_name, rate=dropout_rate)(x)
    if output_reshape is not None:
        reshape_layer_name = "_".join([name, "reshape"])
        if name_ext is not None:
            reshape_layer_name = "_".join([reshape_layer_name, name_ext])
        x = keras.layers.Reshape(
            name=reshape_layer_name,
            target_shape=output_reshape[1:],
        )(x)
    return x


def convolutional_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "convolution",
    name_ext: Optional[str] = None,
    conv_transpose: bool = False,
    **kwargs,
) -> keras.layers.Layer:
    assert len(x.shape) <= 5, "x layer shape should have 5 or less dimensions"
    conv_dim = len(x.shape) - 2
    conv_type = "transpose" if conv_transpose else ""
    conv_name = "_".join([name, "convolution" + str(conv_dim) + "d" + conv_type])
    if name_ext is not None:
        conv_name = "_".join([conv_name, name_ext])
    if conv_dim == 1:
        conv_layer = keras.layers.Conv1D
    elif conv_dim == 2:
        conv_layer = keras.layers.Conv2D
    else:
        conv_layer = keras.layers.Conv3D
    x = conv_layer(
        name=conv_name,
        **kwargs,
    )(x)
    return x


def pooling_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "pooling",
    name_ext: Optional[str] = None,
    pooling: Union[str, keras.layers.Layer] = "max",
    **kwargs,
) -> keras.layers.Layer:
    assert len(x.shape) <= 5, "x layer shape should have 5 or less dimensions"
    pool_dim = len(x.shape) - 2
    pooling_name = "_".join([name, "pooling" + str(pool_dim) + "d"])
    if name_ext is not None:
        pooling_name = "_".join([pooling_name, name_ext])
    if isinstance(pooling, keras.layers.Layer):
        x = pooling(
            name=pooling_name,
            **kwargs,
        )(x)
        return x
    available_max_pooling = ["max", "avg", "global"]
    if pooling not in available_max_pooling:
        raise ValueError(
            f"Unknown pooling type {pooling}. Available pooling: {available_max_pooling}"
        )
    pooling_layer = None
    if pool_dim == 1:
        if pooling == "max":
            pooling_layer = keras.layers.MaxPooling1D
        elif pooling == "avg":
            pooling_layer = keras.layers.AveragePooling1D
        elif pooling == "global":
            pooling_layer = keras.layers.GlobalAveragePooling1D
    elif pool_dim == 2:
        if pooling == "max":
            pooling_layer = keras.layers.MaxPooling2D
        elif pooling == "avg":
            pooling_layer = keras.layers.AveragePooling2D
        elif pooling == "global":
            pooling_layer = keras.layers.GlobalAveragePooling2D
    else:
        if pooling == "max":
            pooling_layer = keras.layers.MaxPooling3D
        elif pooling == "avg":
            pooling_layer = keras.layers.AveragePooling3D
        elif pooling == "global":
            pooling_layer = keras.layers.GlobalAveragePooling3D
    x = pooling_layer(
        name=pooling_name,
        **kwargs,
    )(x)
    return x


def _get_pooling_dim(x: keras.layers.Layer) -> int:
    return len(x.shape) - 2


def self_attention_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "self_attention",
    name_ext: Optional[str] = None,
    sequential_axis: int = 1,
    return_attention_scores: bool = False,
    use_causal_mask: bool = False,
    **kwargs,
) -> keras.layers.Layer:
    sequential_permutation_name = "_".join(
        [name, "sequential_permutation_" + str(sequential_axis)]
    )
    x = sequential_permutation(
        x=x,
        name=sequential_permutation_name,
        name_ext=name_ext,
        sequential_axis=sequential_axis,
    )
    attention_layer_name = "_".join([name, "multi_head_attention"])
    if name_ext is not None:
        attention_layer_name = "_".join([attention_layer_name, name_ext])
    attention_layer = keras.layers.MultiHeadAttention(
        name=attention_layer_name,
        **kwargs,
    )
    if return_attention_scores:
        attention_layer.build(
            query_shape=tf.keras.backend.int_shape(x),
            value_shape=tf.keras.backend.int_shape(x),
        )
        x = attention_layer.call(
            query=x,
            value=x,
            key=x,
            use_causal_mask=use_causal_mask,
            return_attention_scores=return_attention_scores,
        )
    else:
        x = attention_layer(
            query=x,
            value=x,
            key=x,
            use_causal_mask=use_causal_mask,
            return_attention_scores=return_attention_scores,
        )
    return x


def sequential_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "sequential",
    name_ext: Optional[str] = None,
    bidirectional: bool = False,
    sequential_axis: int = 1,
    **kwargs,
) -> keras.layers.Layer:
    sequential_permutation_name = "_".join(
        [name, "sequential_permutation_" + str(sequential_axis)]
    )
    x = sequential_permutation(
        x=x,
        name=sequential_permutation_name,
        name_ext=name_ext,
        sequential_axis=sequential_axis,
    )
    gru_layer_name = "_".join([name, "gru"])
    if name_ext is not None:
        gru_layer_name = "_".join([gru_layer_name, name_ext])
    if bidirectional:
        x = keras.layers.Bidirectional(keras.layers.GRU(name=gru_layer_name, **kwargs))(
            x
        )
    else:
        x = keras.layers.GRU(name=gru_layer_name, **kwargs)(x)
    return x


def dense_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "dense",
    name_ext: Optional[str] = None,
    return_sequences: bool = False,
    **kwargs,
) -> keras.layers.Layer:
    input_shape = x.shape
    if (
        len(input_shape) > 2
        and all([shape is not None for shape in input_shape[1:]])
        and not return_sequences
    ):
        flatten_layer_name = "_".join([name, "pre", "dense", "flatten"])
        if name_ext is not None:
            flatten_layer_name = "_".join([flatten_layer_name, name_ext])
        x = keras.layers.Flatten(
            name=flatten_layer_name,
        )(x)
    dense_layer_name = "_".join([name, "dense"])
    if name_ext is not None:
        dense_layer_name = "_".join([dense_layer_name, name_ext])
    x = keras.layers.Dense(name=dense_layer_name, **kwargs)(x)
    return x


def bn_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer], name: str, name_ext: str, **kwargs
) -> keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2 and all([shape is not None for shape in input_shape[1:]]):
        output_reshape = input_shape
        flatten_layer_name = "_".join([name, "pre", "bn", "flatten"])
        if name_ext is not None:
            flatten_layer_name = "_".join([flatten_layer_name, name_ext])
        x = keras.layers.Flatten(name=flatten_layer_name)(x)
    bn_layer_name = "_".join([name, "batch_normalization"])
    if name_ext is not None:
        bn_layer_name = "_".join([bn_layer_name, name_ext])
    x = keras.layers.BatchNormalization(
        name=bn_layer_name,
        **kwargs,
    )(x)
    if output_reshape is not None:
        reshape_layer_name = "_".join([name, "post", "bn", "reshape"])
        if name_ext is not None:
            reshape_layer_name = "_".join([reshape_layer_name, name_ext])
        x = keras.layers.Reshape(
            name=reshape_layer_name,
            target_shape=output_reshape[1:],
        )(x)
    return x


def ln_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer], name: str, name_ext: str, **kwargs
) -> keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2 and all([shape is not None for shape in input_shape[1:]]):
        output_reshape = input_shape
        flatten_layer_name = "_".join([name, "pre", "ln", "flatten"])
        if name_ext is not None:
            flatten_layer_name = "_".join([flatten_layer_name, name_ext])
        x = keras.layers.Flatten(name=flatten_layer_name)(x)
    ln_layer_name = "_".join([name, "layer_normalization"])
    if name_ext is not None:
        ln_layer_name = "_".join([ln_layer_name, name_ext])
    x = keras.layers.LayerNormalization(
        name=ln_layer_name,
        **kwargs,
    )(x)
    if output_reshape is not None:
        reshape_layer_name = "_".join([name, "post", "ln", "reshape"])
        if name_ext is not None:
            reshape_layer_name = "_".join([reshape_layer_name, name_ext])
        x = keras.layers.Reshape(
            name=reshape_layer_name,
            target_shape=output_reshape[1:],
        )(x)
    return x


def activation_layer_constructor(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "activation",
    name_ext: Optional[str] = None,
    activation: str = "linear",
    **reg_kwargs,
) -> keras.layers.Layer:
    input_shape = x.shape
    output_reshape = None
    if len(input_shape) > 2 and all([shape is not None for shape in input_shape[1:]]):
        output_reshape = input_shape
        flatten_layer_name = "_".join([name, "pre", "activation", "flatten"])
        if name_ext is not None:
            flatten_layer_name = "_".join([flatten_layer_name, name_ext])
        x = keras.layers.Flatten(name=flatten_layer_name)(x)
    activation_layer_name = "_".join(
        [name, activation if isinstance(activation, str) else "activation"],
    )
    if name_ext is not None:
        activation_layer_name = "_".join([activation_layer_name, name_ext])
    if activation == "leakyrelu":
        x = keras.layers.LeakyReLU(name=activation_layer_name)(x)
    elif activation == "prelu":
        x = keras.layers.PReLU(name=activation_layer_name, **reg_kwargs)(x)
    elif activation == "softmax":
        x = keras.layers.Softmax(name=activation_layer_name, dtype="float32")(x)
    else:
        x = keras.layers.Activation(name=activation_layer_name, activation=activation)(
            x
        )
    if output_reshape is not None:
        reshape_layer_name = "_".join([name, "post", "activation", "reshape"])
        if name_ext is not None:
            reshape_layer_name = "_".join([reshape_layer_name, name_ext])
        x = keras.layers.Reshape(
            name=reshape_layer_name,
            target_shape=output_reshape[1:],
        )(x)
    return x


def sequential_permutation(
    x: Union[tf.Tensor, keras.layers.Layer],
    name: str = "permutation",
    name_ext: Optional[str] = None,
    sequential_axis: int = 1,
) -> keras.layers.Layer:
    input_shape = x.shape
    sequential_axis_ = list(range(len(input_shape)))[sequential_axis]
    if sequential_axis_ != 1:
        permutation = tuple(
            [sequential_axis_]
            + [i for i in range(1, len(input_shape[1:])) if i != sequential_axis_],
        )
        permute_layer_name = "_".join([name, "permutation"])
        if name_ext is not None:
            permute_layer_name = "_".join([permute_layer_name, name_ext])
        x = keras.layers.Permute(name=permute_layer_name, dims=permutation)(x)
    else:
        permutation = list(range(1, len(input_shape)))
    if len(input_shape) > 2 and all([shape is not None for shape in input_shape[1:]]):
        permutation_dimensions = [input_shape[i] for i in permutation[1:]]
        reshape_layer_name = "_".join([name, "reshape"])
        if name_ext is not None:
            reshape_layer_name = "_".join([reshape_layer_name, name_ext])
        x = keras.layers.Reshape(
            name=reshape_layer_name,
            target_shape=(
                input_shape[permutation[0]],
                int(np.prod(permutation_dimensions)),
            ),
        )(x)
    return x


def identity(x: keras.layers.Layer) -> keras.layers.Layer:
    """Identity layer."""
    return x



import json
from typing import Dict, Tuple, Union

import keras
import numpy as np
import tensorflow as tf
from numpy.typing import NDArray

from airontools.constructors.layers import layer_constructor
from airontools.constructors.models.model import Model
from airontools.on_the_fly.hyper_design_dropout_rate import HyperDesignDropoutRate


class Sampling(keras.layers.Layer):
    """Uses (z_mean, z_log_var) to sample z, the vector encoding a digit."""

    def call(self, inputs, **kwargs) -> tf.Tensor:
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.random.normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


class VAE(Model, keras.models.Model):
    """Variational AutoEncoder model.
    For more information refer to this paper: https://arxiv.org/pdf/1312.6114.pdf"""

    def __init__(
        self,
        input_shape: Tuple[int],
        model_name: str = "VAE",
        output_activation: str = "linear",
        latent_dim: int = 3,
        **kwargs,
    ):
        Model.__init__(self)
        keras.models.Model.__init__(self)

        # Loss tracker
        self.loss_tracker = keras.metrics.Mean(name="loss")

        # Encoder
        encoder_inputs = keras.layers.Input(shape=input_shape)
        encoder = layer_constructor(
            encoder_inputs,
            units=latent_dim,
            name=f"{model_name}_encoder",
            **kwargs,
        )
        self.z_mean = layer_constructor(
            encoder,
            units=latent_dim,
            name=f"{model_name}_z_mean",
            **kwargs,
        )
        self.z_log_var = layer_constructor(
            encoder,
            units=latent_dim,
            name=f"{model_name}_z_log_var",
            **kwargs,
        )
        self.encoder = keras.models.Model(
            inputs=encoder_inputs,
            outputs=[self.z_mean, self.z_log_var],
            name=f"{model_name}_encoder",
        )

        # Z
        z_mean = keras.layers.Input(shape=(latent_dim,))
        z_log_var = keras.layers.Input(shape=(latent_dim,))
        z = Sampling(name=f"{model_name}_sampling")([z_mean, z_log_var])
        self.z = keras.models.Model(
            inputs=[z_mean, z_log_var],
            outputs=z,
            name=f"{model_name}_z",
        )

        # Decoder
        decoder_inputs = keras.layers.Input(shape=(latent_dim,))
        decoder = layer_constructor(
            decoder_inputs,
            units=input_shape[0],
            name=f"{model_name}_decoder",
            activation=output_activation,
            **kwargs,
        )
        self.decoder = keras.models.Model(
            inputs=decoder_inputs,
            outputs=decoder,
            name=f"{model_name}_decoder",
        )
        self.encoder(encoder_inputs)
        self.z(self.encoder(encoder_inputs))
        self.decoder(self.z(self.encoder(encoder_inputs)))

        # AE
        self._model = keras.models.Model(
            inputs=encoder_inputs,
            outputs=self.decoder(self.z(self.encoder(encoder_inputs))),
            name=model_name,
        )

        # Hyper design on the fly
        self.hyper_design_dropout_rate = HyperDesignDropoutRate(model=self._model)

    def compile(self, *args, **kwargs) -> None:
        """Compile model."""
        keras.models.Model.compile(self, *args, **kwargs)

    def fit(self, *args, **kwargs) -> None:
        """Fit."""
        keras.models.Model.fit(self, *args, **kwargs)

    def evaluate(
        self, x: Union[NDArray[float], tf.Tensor], **kwargs
    ) -> Dict[str, tf.Tensor]:
        """Evaluate model."""
        z_mean, z_log_var = self.encoder(x)
        z = self.z([z_mean, z_log_var])
        loss = self._loss_evaluation(
            inputs=x,
            z_mean=z_mean,
            z_log_var=z_log_var,
            z=z,
        )
        return {"loss": loss}

    def predict(self, *args, **kwargs) -> NDArray[float]:
        """Predict."""
        return keras.models.Model.predict(self, *args, **kwargs)

    def save_weights(self, path: str) -> None:
        """Save model weights."""
        with open(path + "_weights", "w") as f:
            json.dump([w.tolist() for w in self._model.get_weights()], f)

    def load_weights(self, path: str) -> None:
        """Load model weights."""
        with open(path + "_weights") as f:
            encoder_weights = [np.array(w) for w in json.load(f)]
        self._model.set_weights(encoder_weights)

    def call(self, inputs, **kwargs) -> tf.Tensor:
        """Call model."""
        z_mean, z_log_var = self.encoder(inputs)
        z = self.z([z_mean, z_log_var])
        loss, reconstructed = self._loss_evaluation(
            inputs=inputs,
            z_mean=z_mean,
            z_log_var=z_log_var,
            z=z,
            return_reconstruction=True,
        )
        self.add_loss(loss)
        return reconstructed

    def summary(self, **kwargs) -> None:
        """Model summary."""
        self._model.summary(**kwargs)

    def _loss_evaluation(
        self,
        inputs: tf.Tensor,
        z_mean: tf.Tensor,
        z_log_var: tf.Tensor,
        z: tf.Tensor,
        return_reconstruction: bool = False,
    ) -> Union[tf.Tensor, Tuple[tf.Tensor, tf.Tensor]]:
        reconstructed = self.decoder(z)
        rec_loss = tf.reduce_mean((inputs - reconstructed) ** 2)
        # Add KL divergence regularization loss.
        kld_loss = -0.5 * tf.reduce_mean(
            z_log_var - tf.square(z_mean) - tf.exp(z_log_var) + 1
        )
        if return_reconstruction:
            return rec_loss + 0.1 * kld_loss, reconstructed
        else:
            return rec_loss + 0.1 * kld_loss



import json
from typing import Dict, List, Tuple

import keras
import numpy as np
import tensorflow as tf
from numpy.typing import NDArray

from airontools.constructors.layers import layer_constructor
from airontools.constructors.models.model import Model
from airontools.on_the_fly.hyper_design_dropout_rate import HyperDesignDropoutRate


class FeedForward(Model, keras.models.Model):
    """Feed forward model."""

    def __init__(
        self,
        input_shape: Tuple[int],
        n_outputs: int,
        model_name: str = "FeedForward",
        output_activation: str = "linear",
        loss: str = "mse",
        **kwargs,
    ):
        Model.__init__(self)
        keras.models.Model.__init__(self)
        assert loss in [
            "categorical_crossentropy",
            "binary_crossentropy",
            "mse",
        ], f"{loss} is not implemented."
        self._loss = loss

        # Loss tracker
        self.loss_tracker = keras.metrics.Mean(name="loss")

        # Encoder
        encoder_inputs = keras.layers.Input(shape=input_shape)
        encoder_outputs = layer_constructor(
            x=encoder_inputs,
            units=n_outputs,
            name=f"{model_name}_encoder",
            activation=output_activation,
            **kwargs,
        )
        self._model = keras.models.Model(
            inputs=encoder_inputs,
            outputs=encoder_outputs,
            name=model_name,
        )

        # Hyper design on the fly
        self.hyper_design_dropout_rate = HyperDesignDropoutRate(model=self._model)

    def compile(self, *args, **kwargs) -> None:
        """Compile model."""
        keras.models.Model.compile(self, *args, **kwargs)

    def fit(self, *args, **kwargs) -> None:
        """Fit model."""
        keras.models.Model.fit(self, *args, **kwargs)

    def evaluate(
        self, x: NDArray[float], y: NDArray[float], **kwargs
    ) -> Dict[str, tf.Tensor]:
        """Evaluate model."""
        if "sample_weight_val" in kwargs.keys():
            sample_weight = kwargs["sample_weight_val"]
        else:
            sample_weight = 1
        return {"loss": self._loss_evaluation(y, self._model(x), sample_weight)}

    def predict(self, *args, **kwargs) -> NDArray[float]:
        """Predict."""
        return keras.models.Model.predict(self, *args, **kwargs)

    def save_weights(self, path: str) -> None:
        """Save model weights."""
        with open(path + "_weights", "w") as f:
            json.dump([w.tolist() for w in self._model.get_weights()], f)

    def load_weights(self, path: str) -> None:
        """Load model weights."""
        with open(path + "_weights") as f:
            encoder_weights = [np.array(w) for w in json.load(f)]
        self._model.set_weights(encoder_weights)

    def train_step(self, data: List[NDArray[float]], **kwargs) -> Dict[str, float]:
        """Train step."""
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            x, y = data
            sample_weight = 1
        with tf.GradientTape() as tape:
            loss = self._loss_evaluation(y, self._model(x), sample_weight)
        grads = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.trainable_weights))
        self.loss_tracker.update_state(loss)
        return {"loss": self.loss_tracker.result()}

    def call(self, inputs, **kwargs) -> tf.Tensor:
        """Call model."""
        return self._model(inputs, **kwargs)

    def summary(self, **kwargs) -> None:
        """Model summary."""
        self._model.summary(**kwargs)

    def _loss_evaluation(
        self, y: tf.Tensor, y_pred: tf.Tensor, sample_weight: tf.Tensor
    ) -> tf.Tensor:
        if self._loss == "categorical_crossentropy":
            loss = keras.metrics.categorical_crossentropy(y, y_pred)
        elif self._loss == "binary_crossentropy":
            loss = keras.metrics.binary_crossentropy(y, y_pred)
        elif self._loss == "mse":
            loss = keras.metrics.mean_squared_error(y, y_pred)
        else:
            loss = tf.zeros_like(y, dtype=tf.float32)
        loss = loss * sample_weight
        loss = tf.reduce_mean(tf.reduce_sum(loss))
        return loss



import keras

from airontools.constructors.layers import CustomDropout


def save_model(
    model: keras.models.Model,
    filepath: str,
    save_entire_model: bool = False,
    file_format: str = "tf",
) -> None:
    """Save a Keras model.

    Parameters:
        model (keras.models.Model): Keras model to save.
        filepath (str): File path to save the model.
        save_entire_model (bool): Whether to save the entire model as a whole.
        file_format (str): The format of the file, which can be either tf or h5.

    """
    # ToDo: challenge whether this function is still relevant.
    assert file_format in ["tf", "h5"]
    if file_format == "tf":
        file_format = ""
    else:
        file_format = "." + file_format
    model.save_weights(filepath=filepath + "_weights" + file_format)
    with open(filepath + "_topology", "w") as json_file:
        json_file.write(model.to_json())
    if save_entire_model:
        model.save(filepath + file_format)


def load_model(
    filepath: str,
    custom_objects: dict = None,
    load_entire_model: bool = False,
    file_format: str = "tf",
) -> keras.models.Model:
    """Load a Keras model.

    Parameters:
        filepath (str): File path to save the model.
        custom_objects (dict): Custom objects, such CustomDropout, to be loaded together with the model.
        load_entire_model (bool): Whether to load the entire model as a whole.
        file_format (str): The format of the file, which can be either tf or h5.

    Returns:
        x (keras.models.Model): A keras model.

    """
    # ToDo: make the addition of custom objects more general
    if custom_objects is None:
        custom_objects = {"CustomDropout": CustomDropout}
    elif (
        isinstance(custom_objects, dict)
        and "CustomDropout" not in custom_objects.keys()
    ):
        custom_objects.update({"CustomDropout": CustomDropout})
    assert file_format in ["tf", "h5"]
    if file_format == "tf":
        file_format = ""
    else:
        file_format = "." + file_format
    if load_entire_model:
        model = keras.models.load_model(filepath + file_format)
    else:
        json_file = open(filepath + "_topology")
        loaded_model_json = json_file.read()
        json_file.close()
        model = keras.models.model_from_json(loaded_model_json, custom_objects)
        model.load_weights(filepath=filepath + "_weights" + file_format)

    return model


def summary(model) -> None:
    """Model summary.

    Parameters:
        model (Model): Model to summarize.
    """
    print("\n")
    print("________________________ Model Summary __________________________")
    print("Main model name: " + model.name)
    print(model.summary())
    print("\n")
    print("_________________ Layers/Sub-Models Summaries ___________________")
    for layer in model.layers:
        print(layer.name)
        try:
            print(layer.summary())
        except:
            pass
    print("\n")



import warnings
from typing import Optional

import keras


def get_latent_model(
    model: keras.models.Model, layer_name: str
) -> Optional[keras.models.Model]:
    """Gets latent model.

    Parameters:
        model (keras.models.Model): Model.
        layer_name (str): Layer name from which to represent the data.

    Returns:
        A keras.models.Model.
    """
    try:
        return keras.models.Model(
            inputs=model._model.inputs,
            outputs=model._model.get_layer(layer_name).outputs,
            name=layer_name + "_model",
        )
    except ValueError:
        outputs = model._model.inputs
        layer_found = False
        for layer in model._model.layers[1:]:
            outputs = layer(outputs)
            if layer.name == layer_name:
                layer_found = True
                break
        if layer_found:
            return keras.models.Model(
                inputs=model._model.inputs,
                outputs=outputs,
                name=layer_name + "_model",
            )
        else:
            warnings.warn("could not find the layer")


def set_precision(precision: str) -> None:
    """Sets variables precision.

    Parameters:
        precision (str): Precision of the variables.
    """
    if "float16" in precision:
        if precision == "mixed_float16":
            policy = keras.mixed_precision.Policy("mixed_float16")
            keras.mixed_precision.set_global_policy(policy)
        else:
            keras.backend.set_floatx("float16")


def get_regularizer(
    l1_value: Optional[float] = None, l2_value: Optional[float] = None
) -> keras.regularizers.Regularizer:
    """Gets a regularizer.

    Parameters:
        l1_value (float): L1 (Lasso) regularization value.
        l2_value (float): L2 (Ridge) regularization value.

    Returns:
        A keras.regularizers.Regularizer.
    """
    if l1_value and l2_value:
        return keras.regularizers.l1_l2(l1=l1_value, l2=l2_value)
    elif l1_value:
        return keras.regularizers.l1(l1=l1_value)
    elif l2_value:
        return keras.regularizers.l2(l2=l2_value)



import abc
from typing import Dict, List, Union

import tensorflow as tf
from numpy.typing import NDArray


class Model(abc.ABC):
    @abc.abstractmethod
    def compile(self, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def fit(self, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def evaluate(self, *args, **kwargs) -> Dict[str, tf.Tensor]:
        pass

    @abc.abstractmethod
    def predict(self, *args, **kwargs) -> Union[List[NDArray[float]], NDArray[float]]:
        pass

    @abc.abstractmethod
    def save_weights(self, *args, **kwargs) -> None:
        pass

    @abc.abstractmethod
    def load_weights(self, *args, **kwargs) -> None:
        pass



from __future__ import annotations

import os

import setuptools

# Release new version steps
# ToDo: crate a script to automate the following commands
# rm -rf *.egg-info/
# rm -rf build/
# rm -rf dist/
# rm requirements.txt
# poetry export --without-hashes --format=requirements.txt > requirements.txt
# poetry run python setup.py bdist_wheel
# poetry run python -m twine upload dist/*

PACKAGE_NAME = "airontools"

with open("README.md") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    required = f.read().splitlines()


setuptools.setup(
    name=PACKAGE_NAME,
    version="0.1.62",
    scripts=[],
    author="Claudi",
    author_email="claudi_ruiz@hotmail.com",
    description="Machine learning tools to complement the AIronSuit package.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AtrejuArtax/airontools",
    packages=setuptools.find_packages(include=[f"{PACKAGE_NAME}*"]),
    install_requires=required,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    license="BSD",
)
